import os
from configuration import get_config
import threading
from datetime import datetime, timedelta
import time
import record
from services import Worklog, log_all


dir_path = os.path.dirname(os.path.realpath(__file__))
artifacts_path = os.path.join(dir_path, "..", "artifacts")


async def execute_update(from_time, to_time):
    app_id = get_config("log_app_name")
    records = await record.get_latest_answers(from_time=from_time, to_time=to_time)
    # records are Pydantic object not json object

    user_records = {}
    for r in records:
        if r.email not in user_records:
            user_records[r.email] = []
        user_records[r.email].append(r)

    worklogs = []
    for user, user_records in user_records.items():
        user_records.sort(key=lambda x: x.stamp)
        last_stamp = datetime.min
        for r in user_records:
            duration = (r.stamp - last_stamp).total_seconds()
            worklogs.append(
                Worklog(
                    stamp=r.stamp,
                    app_id=app_id,
                    user_id=r.email,
                    disability=r.disability,
                    amount=r.diffamount,
                    duration=duration if duration < 600 else 0,
                    score=-1
                )
            )
            last_stamp = r.stamp

    print("Total records", len(worklogs))
    if len(worklogs) > 0:
        return await log_all(worklogs)
    return True


thread_running = True


async def run_update(update_period_in_seconds):
    print("Converter update thread is starting...")

    last_update_stamp_path = os.path.join(artifacts_path, "last_update_stamp")

    if not os.path.exists(last_update_stamp_path):
        last_converter_update = datetime.utcnow() - timedelta(days=350)
    else:
        with open(last_update_stamp_path, "r") as file:
            last_converter_update = datetime.fromtimestamp(float(file.read()))

    while thread_running:
        now = datetime.utcnow()
        if (now - last_converter_update).total_seconds() > update_period_in_seconds:
            print("computing logs...", now)
            if await execute_update(last_converter_update, now):
                last_converter_update = now
                with open(last_update_stamp_path, "w") as file:
                    file.write(str(last_converter_update.timestamp()))
        time.sleep(60)


the_thread = None


def set_update_schedule(update_period_in_seconds=60):
    global the_thread, thread_running
    terminate_schedule()
    the_thread = threading.Thread(target=run_update, args=(update_period_in_seconds,))
    thread_running = True
    the_thread.start()


def terminate_schedule():
    global thread_running
    thread_running = False
    if the_thread:
        the_thread.join()


if __name__ == '__main__':
    import asyncio
    asyncio.run(run_update(3600))



from pymongo import MongoClient, DESCENDING, ASCENDING


class Mongo:
    def __init__(self, url, database):
        self.client = MongoClient(url)
        self.database = database

    def get_session(self):
        return self.client[self.database]



import json
import os


class Configuration:

    def __init__(self):
        self.data = {
            "test": True,
            "secret_key": "1234"
        }

    def load(self, path):
        with open(path, "rb") as file:
            self.data = json.load(file)

    def __getitem__(self, key):
        return self.data[key] if key in self.data else None


config_object = Configuration()


def get_config(key: str):
    return config_object[key]


def get_config_object():
    return config_object


def initialize(project_root):
    config_file_path = os.path.join(project_root, "configurations", "config.json")

    if os.path.exists(config_file_path):
        config_object.load(config_file_path)


if __name__ == '__main__':
    initialize(os.path.join(".."))
    print(get_config("data_path"))



from fastapi import Depends, HTTPException, status, Form, Header, APIRouter
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from typing import Optional
import jwt
from pydantic import BaseModel
from datetime import datetime, timedelta
from .configuration import get_config
from .services import *


# to get a string like this run:
# openssl rand -hex 32
ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 24 * 60
REFRESH_TOKEN_EXPIRE_MINUTES = 24 * 60 * 5


def create_access_token(data: dict, expires_delta: Optional[timedelta] = timedelta(minutes=15)):
    SECRET_KEY = get_config("secret_key")
    expire = datetime.utcnow() + expires_delta
    to_encode = data.copy()
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


def create_refresh_token(data: dict, expires_delta: Optional[timedelta] = timedelta(minutes=30)):
    return create_access_token(data=data, expires_delta=expires_delta)


def create_tokens(userdata: dict):
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        data=userdata, expires_delta=access_token_expires
    )
    refresh_token = create_refresh_token(
        data=userdata, expires_delta=timedelta(minutes=REFRESH_TOKEN_EXPIRE_MINUTES)
    )
    return Token(access_token=access_token, refresh_token=refresh_token, token_type="bearer")


class User(BaseModel):
    email: str
    disability: int


credentials_exception = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail="Could not validate credentials",
    headers={"Authorization": "Bearer"},
)


def decode_token(token: str):
    SECRET_KEY = get_config("secret_key")
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        email: str = payload.get("email")
        exp = payload.get("exp")
        disability = payload.get("disability")
        expired = datetime.now() > datetime.fromtimestamp(exp)
        if expired:
            raise jwt.ExpiredSignatureError()
        if email is None:
            raise credentials_exception
    except jwt.DecodeError:
        raise credentials_exception
    except jwt.ExpiredSignatureError:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Token is expired.",
            headers={"WWW-Authenticate": "Bearer"},
        )

    return email, disability


class Token(BaseModel):
    access_token: str
    refresh_token: str
    token_type: str


def initialize(app, security_prefix="", get_user=None):

    router = APIRouter(prefix=security_prefix, tags=[security_prefix[1:] if security_prefix.startswith("/") else security_prefix])

    @router.post("/exchange", response_model=Token)
    async def exchange_access_token(token: str = Form(...)):
        """
            Login as a collab user
        """
        bypass_security = get_config("test")
        if bypass_security:
            userdata = {'email': "testuser@vulcan", 'disability': 0}
        else:
            result, userdata = await login_linkage(token)
            if not result:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="Incorrect username or password",
                    headers={"WWW-Authenticate": "Bearer"},
                )

        return create_tokens(userdata)

    @router.post('/refresh', response_model=Token)
    async def refresh_token(refresher: str = Form(...)):

        email, disability = decode_token(refresher)

        return create_tokens({"email": email, "disability": disability})

    @router.post("/token", response_model=Token)
    @router.post("/admin", response_model=Token)
    async def login_for_admin_token(form_data: OAuth2PasswordRequestForm = Depends()):
        """
            Login as a value performanceManager
        """
        bypass_security = get_config("test")
        if bypass_security:
            userdata = {'email': "testadmin@vulcan", 'data': True}
        else:
            result, userdata = await admin_login(form_data.username, form_data.password)
            if not result:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="Incorrect username or password",
                    headers={"WWW-Authenticate": "Bearer"},
                )
            userdata["data"] = True

        return create_tokens(userdata)

    @router.post("/login", response_model=Token)
    async def login(form_data: OAuth2PasswordRequestForm = Depends()):
        """
            custom user login (required parameter get_user)
        """
        bypass_security = get_config("test")
        if bypass_security:
            userdata = {'email': "testuser@vulcan", 'data': False}
            return create_tokens(userdata)
        elif get_user is not None:
            result, userdata = await get_user(form_data.username, form_data.password)
            if result:
                userdata["data"] = False
                return create_tokens(userdata)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )

    app.include_router(router)

    async def get_active_current_user(Authorization: str = Header(None)):
        bypass_security = get_config("test")
        if Authorization is None:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Token is missing.",
                headers={"WWW-Authenticate": "Bearer"},
            )

        token = Authorization[len("Bearer "):]

        if bypass_security:
            return User(email="testuser@vulcan", disability=0)

        email, disability = decode_token(token)

        user = User(email=email, disability=disability)
        return user

    oauth2_scheme = OAuth2PasswordBearer(tokenUrl=security_prefix + "/token")

    async def check_is_admin(token: str = Depends(oauth2_scheme)):
        SECRET_KEY = get_config("secret_key")
        try:
            payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
            email: str = payload.get("email")
            exp = payload.get("exp")
            data = payload.get("data")
            expired = datetime.now() > datetime.fromtimestamp(exp)
            if email is None or expired or data is None:
                raise credentials_exception
        except jwt.DecodeError:
            raise credentials_exception

        return True

    async def get_admin(token: str = Depends(oauth2_scheme)):
        SECRET_KEY = get_config("secret_key")
        try:
            payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
            email: str = payload.get("email")
            exp = payload.get("exp")
            data = payload.get("data")
            expired = datetime.now() > datetime.fromtimestamp(exp)
            if email is None or expired or data is None:
                raise credentials_exception
        except jwt.DecodeError:
            raise credentials_exception

        return User(email=email, disability=0)

    return get_active_current_user, check_is_admin, get_admin



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.types import TypeDecorator, DateTime
from datetime import datetime
import pytz

Base = declarative_base()


def create_database(path="vulcan@localhost:5432/overlord"):
    SQLALCHEMY_DATABASE_URL = "postgresql://" + path
    engine = create_engine(
        SQLALCHEMY_DATABASE_URL, pool_size=128, max_overflow=32
    )
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
    return engine, SessionLocal


class Postgres:
    def __init__(self, url):
        self.engine, self.session = create_database(url)
        Base.metadata.create_all(bind=self.engine)

    def get_session(self):
        return self.session()


class TimeStamp(TypeDecorator):
    impl = DateTime
    LOCAL_TIMEZONE = datetime.utcnow().astimezone().tzinfo
    cache_ok = True

    def process_bind_param(self, value: datetime, dialect):
        if value.tzinfo is None:
            value = value.astimezone(self.LOCAL_TIMEZONE)

        return value.astimezone(pytz.utc)

    def process_result_value(self, value, dialect):
        if value.tzinfo is None:
            return value.replace(tzinfo=pytz.utc)



from .configuration import initialize as init_config, get_config
from .security import initialize as init_security, User
from .services import *
from .database import initialize as init_db, terminate_connections, Base, TimeStamp


def initialize(project_root, fast_api_app=None, security_prefix="", get_user=None):
    init_config(project_root)
    init_db(configuration.get_config_object())
    if fast_api_app is not None:
        return init_security(fast_api_app, security_prefix, get_user)
    return None


def terminate():
    terminate_connections()



from sshtunnel import SSHTunnelForwarder
import os


dir_path = os.path.dirname(os.path.realpath(__file__))


class Tunnel_Session:
    def __init__(self, credentials):
        self.server = SSHTunnelForwarder(
            credentials["host"],
            ssh_username=credentials["username"],
            ssh_pkey=os.path.join(dir_path, "..", "..", "artifacts", credentials["pkey"]),
            remote_bind_address=(credentials["bind_host"], credentials["bind_port"])
        )
        self.server.start()
        print("Tunneling start")

    def get_bind_addresses(self):
        return '127.0.0.1', self.server.local_bind_port

    def disconnect(self):
        self.server.stop()



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="vulcan-app",
    version="1.3.2",
    author="Chatavut Viriyasuthee",
    author_email="chatavut@lab.ai",
    description="Vulcan's web application server.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vulcan-coalition/vulcan-app.git",
    packages=["vulcan_app", "vulcan_app.database"],
    package_data={},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=[
        "aiofiles",
        "fastapi",
        "httpx==0.23.3",
        "pydantic",
        "PyJWT",
        "pytz",
        "requests",
        "requests-async",
        "SQLAlchemy",
        "PyDrive2",
        "pymongo",
        "boto3",
        "asyncpg",
        "shortuuid",
        "python-multipart",
        "sshtunnel==0.4.0"
    ]
)



import requests
from .configuration import get_config
from datetime import datetime
from pydantic import BaseModel


async def login_linkage(token: str):
    address = get_config("authentication_service_address")
    if address:
        headers = {
            "Authorization": "Bearer " + token
        }
        response = requests.get(address, headers=headers)
        try:
            response.raise_for_status()
            data = response.json()
            return True, {'email': data['email'], 'disability': data['disability']}
        except requests.exceptions.HTTPError as exc:
            print("Login failed", exc)
            return False, None
    return True, {'email': "testuser@vulcan", 'disability': 0}


async def admin_login(username: str, password: str):
    login_address = get_config("admin_authentication_service_address")
    user_address = get_config("admin_verify_service_address")
    if login_address and user_address:

        r = requests.post(login_address, data={'username': username, 'password': password})
        if r.status_code < 400:
            data = r.json()
            headers = {
                "Authorization": "Bearer " + data["access_token"]
            }
            response = requests.get(user_address, headers=headers)
            try:
                response.raise_for_status()
                data = response.json()
                if "performanceManager" in data["claims"]:
                    return True, {'email': data['email']}
                return False, None
            except requests.exceptions.HTTPError as exc:
                print("Login failed", exc)
                return False, None
        else:
            return False, None

    return False, None


class Worklog(BaseModel):
    stamp: datetime
    app_id: str
    user_id: str
    disability: int
    amount: float
    duration: float
    score: float

    def serialize(self):
        d = self.dict()
        d["stamp"] = d["stamp"].strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        return d


async def log_all(worklogs):
    address = get_config('logging_service_address')
    if address:
        response = requests.post(address, json=[w.serialize() for w in worklogs])
        try:
            response.raise_for_status()
            return True
        except requests.exceptions.HTTPError as exc:
            print("Logging failed", exc)
            return False
    return True


if __name__ == '__main__':
    pass



from .tunnel import Tunnel_Session
from .postgre import Postgres, Base, TimeStamp
from .mongo import Mongo

database = None
ssh_tunnel = None


def initialize(config, override_address=None, override_port=None):
    global database, ssh_tunnel

    override_address = None
    override_port = None
    if config["ssh_tunnel"] is not None:
        credentials = config["ssh_tunnel"]
        if credentials is not None:
            ssh_tunnel = Tunnel_Session(credentials)
            override_address, override_port = ssh_tunnel.get_bind_addresses()

    if config["postgre_connection"] is not None:
        connection_str = config["postgre_connection"]
        if not isinstance(connection_str, str):
            username = connection_str["username"]
            password = connection_str["password"]
            hostname = connection_str["hostname"] if override_address is None else override_address
            port = connection_str["port"] if override_port is None else str(override_port)
            db_name = connection_str["db_name"]
            connection_str = username + ":" + password + "@" + hostname + ":" + port + "/" + db_name
        database = Postgres(connection_str)
    elif config["mongo_connection"] is not None:
        url = config["mongo_connection"]
        table = config["mongo_database"]
        database = Mongo(url, table)


def get_session():
    # for postgres
    return database.get_session()


def terminate_connections():
    if ssh_tunnel is not None:
        ssh_tunnel.disconnect()
